from .envi_file_interface import (
    read_spectral_envi_file_context,
    read_measurement_envi_file_context,
)

__all__ = [
    "read_spectral_envi_file_context",
    "read_measurement_envi_file_context",
]



from .geotransform_model import GeotransformModel, PointModel
from .gcp_model import GCPGroup, GroundControlPoint, ImageOffset
from .bounding_box_model import BoundingBoxModel


__all__ = [
    "GeotransformModel",
    "GCPGroup",
    "GroundControlPoint",
    "ImageOffset",
    "PointModel",
    "BoundingBoxModel",
]



# Built-ins
from dataclasses import dataclass
from typing import Literal

# Dependencies
import numpy as np
import xarray as xr

# Local imports
from cubio.types import CubeArrayFormat


@dataclass
class CubeSize:
    nrows: int
    ncolumns: int
    nbands: int

    def as_tuple(self, interleave: CubeArrayFormat) -> tuple[int, int, int]:
        if interleave == "BIL":
            return (self.nrows, self.nbands, self.ncolumns)
        elif interleave == "BIP":
            return (self.nrows, self.ncolumns, self.nbands)
        elif interleave == "BSQ":
            return (self.nbands, self.ncolumns, self.nrows)


def get_cube_size(
    arr: np.ndarray | np.memmap | xr.DataArray, format: CubeArrayFormat
) -> CubeSize:
    if format == "BIL":
        lines, bands, samples = arr.shape
    elif format == "BIP":
        lines, samples, bands = arr.shape
    elif format == "BSQ":
        bands, samples, lines = arr.shape

    return CubeSize(lines, samples, bands)


def transpose_cube(
    src: CubeArrayFormat,
    dst: CubeArrayFormat | Literal["RASTERIO"],
    arr: xr.DataArray,
):
    if src == "BIL":
        if dst == "BIL":
            return arr
        elif dst == "BIP":
            return arr.transpose(*(arr.dims[0], arr.dims[2], arr.dims[1]))
        elif dst == "BSQ":
            return arr.transpose(*(arr.dims[1], arr.dims[2], arr.dims[0]))
        elif dst == "RASTERIO":
            return arr.transpose(*(arr.dims[1], arr.dims[0], arr.dims[2]))
    elif src == "BIP":
        if dst == "BIL":
            return arr.transpose(*(arr.dims[0], arr.dims[2], arr.dims[1]))
        elif dst == "BIP":
            return arr
        elif dst == "BSQ":
            return arr.transpose(*(arr.dims[2], arr.dims[1], arr.dims[0]))
        elif dst == "RASTERIO":
            return arr.transpose(*(arr.dims[2], arr.dims[0], arr.dims[1]))
    elif src == "BSQ":
        if dst == "BIL":
            return arr.transpose(*(arr.dims[2], arr.dims[0], arr.dims[1]))
        elif dst == "BIP":
            return arr.transpose(*(arr.dims[2], arr.dims[1], arr.dims[0]))
        elif dst == "BSQ":
            return arr
        elif dst == "RASTERIO":
            return arr.transpose(*(arr.dims[0], arr.dims[2], arr.dims[1]))



# Built-Ins
import re
from pathlib import Path
from typing_extensions import Self
from uuid import UUID, uuid4

# Dependencies
import numpy as np
from pydantic import BaseModel, Field
import xarray as xr


class GroundControlPoint(BaseModel):
    pixel_row: float = Field(
        ..., description="Pixel row coordinate of the GCP"
    )
    pixel_column: float = Field(
        ..., description="Pixel column coordinate of the GCP"
    )
    map_x: float = Field(..., description="Map x coordinate of the GCP")
    map_y: float = Field(..., description="Map y coordinate of the GCP")
    id: UUID = Field(
        default_factory=uuid4, description="Unique identifier for the GCP"
    )


class ImageOffset(BaseModel):
    height: int = Field(..., description="Height of the image")
    width: int = Field(..., description="Width of the image")
    row: int = Field(..., description="Row offset for the image")
    column: int = Field(..., description="Column offset for the image")

    @property
    def row_slice(self) -> slice:
        return slice(self.row, self.row + self.height - 1)

    @property
    def col_slice(self) -> slice:
        return slice(self.column, self.column + self.width - 1)

    def crop_image(
        self, base_image: np.ndarray | xr.DataArray
    ) -> np.ndarray | xr.DataArray:
        """
        Extract the offset image from a corresponding base image.
        """
        if (base_image.shape[0] < self.row + self.height) or (
            base_image.shape[1] < self.column + self.width
        ):
            raise ValueError(
                "Input image (size:"
                f"{base_image.shape[0], base_image.shape[1]}) is too small for"
                " the offset size: "
                f"({self.row + self.height, self.column + self.width})."
            )

        return base_image[self.row_slice, self.col_slice]

    def as_tuple(self) -> tuple[int, int, int, int]:
        """(Row Offset, Column Offset, Height, Width)"""
        return (self.row, self.column, self.height, self.width)


class GCPGroup(BaseModel):
    offset: ImageOffset
    gcp_list: list[GroundControlPoint]

    @classmethod
    def from_txt(cls, file_path: Path | str, headerrows: int = 7) -> Self:
        data = np.loadtxt(
            file_path, dtype="<U36", skiprows=headerrows, delimiter=","
        )
        good_cols = []
        for i in range(data.shape[1]):
            if np.all(data[:, i] == " "):
                continue
            good_cols.append(i)

        data = data[:, np.array(good_cols)]

        gcp_list: list[GroundControlPoint] = []
        for idx in range(data.shape[0]):
            row = data[idx, :]
            gcp = GroundControlPoint(
                pixel_row=row[1],
                pixel_column=row[2],
                map_x=row[3],
                map_y=row[4],
            )
            gcp_list.append(gcp)

        hdr_pattern = re.compile(r"Source for Target Image:[\s\S]*?ID")
        with open(file_path, "r") as f:
            file_content = f.read()
            test = re.search(hdr_pattern, file_content)
        if test is None:
            raise ValueError(f"Invalid File Header: {file_content}")
        header = file_content[slice(*test.span())]
        header_lines = header.split("\n")

        search_list = [
            re.search(r"([\s\S]*?):([\s\S]*)", i) for i in header_lines[:-1]
        ]
        header_dict: dict[str, int] = {}
        for item in search_list:
            if item is not None:
                g = item.groups()
                try:
                    header_dict[g[0]] = int(g[1])
                except ValueError:
                    continue

        offset = ImageOffset(
            height=header_dict["Target Image Height"],
            width=header_dict["Target Image Width"],
            row=header_dict["Row Offset"],
            column=header_dict["Column Offset"],
        )

        return cls(
            offset=offset,
            gcp_list=gcp_list,
        )

    @classmethod
    def from_gcps_file(cls, fp: Path | str) -> Self:
        with open(fp, "r") as f:
            json = f.read()
        return cls.model_validate_json(json)

    @property
    def ngcp(self) -> int:
        return len(self.gcp_list)

    @property
    def row_pixels(self) -> np.ndarray:
        return np.array([gcp.pixel_row for gcp in self.gcp_list])

    @property
    def col_pixels(self) -> np.ndarray:
        return np.array([gcp.pixel_column for gcp in self.gcp_list])

    @property
    def map_x(self) -> np.ndarray:
        return np.array([gcp.map_x for gcp in self.gcp_list])

    @property
    def map_y(self) -> np.ndarray:
        return np.array([gcp.map_y for gcp in self.gcp_list])

    def add_gcp(self, gcp: GroundControlPoint):
        self.gcp_list.append(gcp)

    def write_json(self, fp: Path | str) -> None:
        with open(Path(fp).with_suffix(".gcps"), "w") as f:
            f.write(self.model_dump_json(indent=2))

    def write_csv(self, fp: Path | str) -> None:
        with open(Path(fp).with_suffix(".csv"), "w") as f:
            f.write("Index, Pixel Row, Pixel Column, Map X, Map Y, ID\n")
            for n, i in enumerate(self.gcp_list):
                f.write(
                    f"{n}, {i.pixel_row}, {i.pixel_column}, {i.map_x},"
                    f" {i.map_y}, {i.id}\n"
                )

    def __add__(self, value: Self) -> Self:
        for i in value.gcp_list:
            self.add_gcp(i)

        return self

    def adjust_offset(self, new_offset: ImageOffset):
        row_diff = self.offset.row - new_offset.row
        col_diff = self.offset.column - new_offset.column
        new_gcp_list: list[GroundControlPoint] = []
        for gcp in self.gcp_list:
            new_gcp = GroundControlPoint(
                pixel_row=gcp.pixel_row + row_diff,
                pixel_column=gcp.pixel_column + col_diff,
                map_x=gcp.map_x,
                map_y=gcp.map_y,
            )
            new_gcp_list.append(new_gcp)
        self.offset = new_offset
        self.gcp_list = new_gcp_list



# Built-ins
from dataclasses import dataclass
from typing import Literal
from typing_extensions import Self

# Dependencies
from pydantic import BaseModel
from affine import Affine  # type: ignore
import numpy as np


class GeographicBoundsError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class PointModel(BaseModel):
    """Representation of a ordered pair"""

    x: float
    y: float

    def astuple(self):
        return (self.x, self.y)


@dataclass
class Bounds:
    """Representation of a bounding box"""

    left: float
    bottom: float
    right: float
    top: float


class GeotransformModel(BaseModel):
    """
    Object representing an affine geotransform matrix.

    Parameters
    ----------
    upperleft: PointModel
        Map coordinates of the upper left point of the raster.
    xres: float
        east-west resolution of the raster in map_unit/pixel.
    row_rotation: float
        Amount of rotation for the rows. Usually 0.
    yres: float
        north-south resolution of the raster in map_unit/pixel.
    col_rotation
        Amount of rotation for the columns. Usually 0.
    """

    upperleft: PointModel
    xres: float
    row_rotation: float
    yres: float
    col_rotation: float

    @classmethod
    def null(cls):
        return cls(
            upperleft=PointModel(x=0, y=0),
            xres=1,
            row_rotation=0,
            yres=1,
            col_rotation=0,
        )

    @classmethod
    def fromgdal(
        cls, gdal_transform: tuple[float, float, float, float, float, float]
    ) -> Self:
        """Creates a GeotransformModel from a GDAL-style geotransform."""
        upper_left_pt = PointModel(x=gdal_transform[0], y=gdal_transform[3])
        return cls(
            upperleft=upper_left_pt,
            xres=gdal_transform[1],
            row_rotation=gdal_transform[2],
            yres=gdal_transform[5],
            col_rotation=gdal_transform[4],
        )

    @classmethod
    def fromaffine(cls, affine_transform: Affine) -> Self:
        upper_left_pt = PointModel(x=affine_transform.c, y=affine_transform.f)
        return cls(
            upperleft=upper_left_pt,
            xres=affine_transform.a,
            row_rotation=affine_transform.b,
            yres=affine_transform.e,
            col_rotation=affine_transform.d,
        )

    @classmethod
    def fromarraysize(
        cls,
        upper_left_latitiude: float,
        upper_left_longitude: float,
        lower_right_latitude: float,
        lower_right_longitude: float,
        height: int,
        width: int,
    ) -> Self:
        upper_left_pt = PointModel(
            x=upper_left_longitude, y=upper_left_latitiude
        )
        lower_right_pt = PointModel(
            x=lower_right_longitude, y=lower_right_latitude
        )
        xres = abs(upper_left_pt.x - lower_right_pt.x) / width
        yres = abs(upper_left_pt.y - lower_right_pt.y) / height
        return cls(
            upperleft=upper_left_pt,
            xres=xres,
            row_rotation=0,
            yres=-yres,
            col_rotation=0,
        )

    def togdal(self) -> tuple[float, float, float, float, float, float]:
        """Returns a GDAL-style geotransform."""
        return (
            self.upperleft.x,
            self.xres,
            self.row_rotation,
            self.upperleft.y,
            self.col_rotation,
            self.yres,
        )

    def toaffine(self) -> Affine:
        return Affine(
            a=self.xres,
            b=self.row_rotation,
            c=self.upperleft.x,
            d=self.col_rotation,
            e=self.yres,
            f=self.upperleft.y,
        )

    def get_bbox(self, height: int, width: int):
        """Given the height and width of a raster, return a bounding box."""
        return Bounds(
            left=self.upperleft.x,
            bottom=self.upperleft.y + height * self.yres,
            right=self.upperleft.x + width * self.xres,
            top=self.upperleft.y,
        )

    def pixel_to_map(
        self,
        xpixel: float,
        ypixel: float,
        convention: Literal["globe", "hemi"] = "hemi",
    ) -> PointModel:
        """Convert a pixel coordinate point to a map coordinate point."""
        xmap = (
            self.upperleft.x + xpixel * self.xres + ypixel * self.row_rotation
        )
        ymap = (
            self.upperleft.y + ypixel * self.yres + xpixel * self.col_rotation
        )

        if convention == "globe":
            if xmap < 0:
                xmap += 360
        return PointModel(x=xmap, y=ymap)

    def map_to_pixel(self, xmap: float, ymap: float) -> PointModel:
        """Convert a map coordinate point to a pixel coordinate point."""
        _scaler = (self.xres * self.yres) - (
            self.col_rotation * self.row_rotation
        )
        xpixel = (
            (self.yres * xmap)
            - (self.upperleft.x * self.yres)
            - ((self.row_rotation * self.yres) / self.xres) * ymap
            + (self.row_rotation * self.upperleft.y)
        ) / _scaler
        ypixel = (
            (self.xres * ymap)
            - (self.upperleft.y * self.xres)
            - ((self.col_rotation * self.xres) / self.yres) * xmap
            + (self.col_rotation * self.upperleft.x)
        ) / _scaler

        if xpixel < 0:
            raise GeographicBoundsError(
                f"{xmap} is beyond the left X bound: {self.upperleft.x}"
            )
        if ypixel < 0:
            raise GeographicBoundsError(
                f"{ymap} is beyond the top bound: {self.upperleft.y}"
            )
        return PointModel(x=xpixel, y=ypixel)

    def generate_coords(
        self, *, width: int, height: int
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Generates coordinate using the given geotransform for an array of size
        xsize by ysize.

        Parameters
        ----------
        xsize : int
            Size of the desired coordinate grid in the x direction (width).
        ysize : int
            Size of the desired coordinate grid in the y direction (height).

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            Tuple of (x-coordinates, y-coordinates).
        """
        xcoords = np.empty(width, dtype=np.float32)
        ycoords = np.empty(height, dtype=np.float32)
        for nx in range(width):
            xcoords[nx] = self.pixel_to_map(xpixel=nx, ypixel=0).x
        for ny in range(height):
            ycoords[ny] = self.pixel_to_map(xpixel=0, ypixel=ny).y
        return xcoords, ycoords



"""
Welcome to `cubio`!
---
### Available Classes
- CubeContext // Metadata for reading in geospatial cubes.
- CubeData // Memory-mapped pointer for performing processing with `xarray`.
"""

from .cube_context import CubeContext

__all__ = ["CubeContext"]



from pathlib import Path

from cubio.cube_context import CubeContext
from cubio.cube_data import CubeData
from cubio.types import (
    CubeArrayFormat,
    cube_array_suffix_map,
    RasterioProfile,
)

import rasterio as rio  # type: ignore


def write_envi(
    dst_fp: Path | str,
    cube_context: CubeContext,
    cube_data: CubeData,
    interleave: CubeArrayFormat,
) -> None:
    prf: RasterioProfile = {
        "height": cube_context.shape.nrows,
        "width": cube_context.shape.ncolumns,
        "count": cube_context.shape.nbands,
        "crs": cube_context.crs,
        "driver": "ENVI",
        "dtype": cube_context.data_type,
        "interleave": interleave,
        "nodata": -999,
        "transform": cube_context.geotransform.toaffine(),
    }
    save_fp = Path(
        dst_fp,
        cube_context.data_filename.with_suffix(
            cube_array_suffix_map[interleave]
        ),
    )
    r_cube = cube_data.transpose_to_rasterio()
    with rio.open(save_fp, "w", **prf) as f:
        f.write(r_cube)
    cube_context.write_envi_hdr(dst=dst_fp, use_image_name=True)



# Built-Ins
from pathlib import Path

# Dependencies
import xarray as xr
import numpy as np

# Local Imports
from cubio.types import suffix_to_format_map, NumpyDType
from cubio.cube_size_tools import CubeSize


def cube_reader(
    fp: Path, size: CubeSize, data_type: NumpyDType
) -> xr.DataArray:
    suff = fp.suffix
    binary_fmt = suffix_to_format_map.get(suff)
    if binary_fmt is not None:
        arr = np.memmap(
            fp, dtype=np.dtype(data_type), shape=size.as_tuple(binary_fmt)
        )
        return xr.DataArray(arr)
    else:
        raise NotImplementedError()



from typing import TypeAlias, Literal, TypeGuard, TypedDict, NotRequired
from enum import StrEnum
from rasterio.crs import CRS  # type: ignore
from affine import Affine  # type: ignore
import numpy as np

CubeArrayFormat: TypeAlias = Literal["BIL", "BIP", "BSQ"]
cube_array_foramts: list[CubeArrayFormat] = ["BIL", "BIP", "BSQ"]
cube_array_suffix_map: dict[CubeArrayFormat, str] = {
    "BIL": ".bil",
    "BIP": ".bip",
    "BSQ": ".bsq",
}
suffix_to_format_map: dict[str, CubeArrayFormat] = {
    v: k for k, v in cube_array_suffix_map.items()
}


def is_valid_cubearrayformat(value: str) -> TypeGuard[CubeArrayFormat]:
    return value in cube_array_foramts


class NumpyDType(StrEnum):
    UINT8 = "uint8"  # unsigned 8-bit integer
    INT16 = "int16"  # signed 16-bit integer
    INT32 = "int32"  # signed 32-bit integer
    FLOAT32 = "float32"  # 32-bit floating point
    FLOAT64 = "float64"  # 64-bit floating point
    CPLX64 = "complex64"  # 64-bit complex number (2 x 32-bit floats)
    CPLX128 = "complex128"  # 128-bit complex number (2 x 64-bit floats)
    UINT16 = "uint16"  # unsigned 16-bit integer
    UINT32 = "uint32"  # unsigned 32-bit integer
    INT64 = "int64"  # signed 64-bit integer
    UINT64 = "uint64"  # unsigned 64-bit integer


NumpyDTypeLiteral: TypeAlias = Literal[
    "UINT8",
    "INT16",
    "INT32",
    "FLOAT32",
    "FLOAT64",
    "CPLX64",
    "CPLX128",
    "UINT16",
    "UINT32",
    "INT64",
    "UINT64",
]

dtype_to_hdr_integer: dict[NumpyDType, int] = {
    NumpyDType.UINT8: 1,
    NumpyDType.INT16: 2,
    NumpyDType.INT32: 3,
    NumpyDType.FLOAT32: 4,
    NumpyDType.FLOAT64: 5,
    NumpyDType.CPLX64: 6,
    NumpyDType.CPLX128: 9,
    NumpyDType.UINT16: 12,
    NumpyDType.UINT32: 13,
    NumpyDType.INT64: 14,
    NumpyDType.UINT64: 15,
}
hdr_integer_to_dtype = {v: k for k, v in dtype_to_hdr_integer.items()}

RasterioDriver: TypeAlias = Literal["ENVI", "GTiff", "ISIS3"]
rasterio_drivers: list[RasterioDriver] = ["ENVI", "GTiff", "ISIS3"]


class RasterioProfile(TypedDict):
    width: int
    height: int
    count: int
    driver: RasterioDriver
    interleave: NotRequired[CubeArrayFormat]
    crs: CRS
    transform: Affine
    dtype: NumpyDType
    nodata: float | int


LabelLike: TypeAlias = np.ndarray | list[float] | list[str]


ImageSuffix: TypeAlias = Literal[
    ".zarr", ".bsq", ".bil", ".bip", ".img", ".hdf5"
]
valid_image_suffixes: list[ImageSuffix] = [
    ".zarr",
    ".bsq",
    ".bil",
    ".bip",
    ".img",
    ".hdf5",
]
image_suffix_priority: dict[ImageSuffix, int] = {
    ".zarr": 0,
    ".bil": 1,
    ".bip": 2,
    ".bsq": 3,
    ".img": 4,
    ".hdf5": 5,
}


def is_valid_image_suffix(value: str) -> TypeGuard[ImageSuffix]:
    return value in valid_image_suffixes



# flake8:noqa

from enum import StrEnum


class GeographicCRS(StrEnum):
    GCS_MOON_2000 = 'GEOGCRS["GCS_Moon_2000", DATUM["D_Moon_2000", ELLIPSOID["Moon_2000_IAU_IAG",1737400,0,LENGTHUNIT["metre",1]]],PRIMEM["Reference_Meridian",0,ANGLEUNIT["degree",0.0174532925199433]],CS[ellipsoidal,2],AXIS["geodetic latitude (Lat)",north,ORDER[1],ANGLEUNIT["degree",0.0174532925199433]],AXIS["geodetic longitude (Lon)",east,ORDER[2],ANGLEUNIT["degree",0.0174532925199433]],USAGE[SCOPE["Not known."],AREA["World."],BBOX[-90,-180,90,180]],ID["ESRI",104903]]'
    WGS84 = 'GEOGCRS["WGS 84",ENSEMBLE["World Geodetic System 1984 ensemble",MEMBER["World Geodetic System 1984 (Transit)"],MEMBER["World Geodetic System 1984 (G730)"],MEMBER["World Geodetic System 1984 (G873)"],MEMBER["World Geodetic System 1984 (G1150)"],MEMBER["World Geodetic System 1984 (G1674)"],MEMBER["World Geodetic System 1984 (G1762)"],MEMBER["World Geodetic System 1984 (G2139)"],MEMBER["World Geodetic System 1984 (G2296)"],ELLIPSOID["WGS 84",6378137,298.257223563,LENGTHUNIT["metre",1]],ENSEMBLEACCURACY[2.0]],PRIMEM["Greenwich",0,ANGLEUNIT["degree",0.0174532925199433]],CS[ellipsoidal,2],AXIS["geodetic latitude (Lat)",north,ORDER[1],ANGLEUNIT["degree",0.0174532925199433]],AXIS["geodetic longitude (Lon)",east,ORDER[2],ANGLEUNIT["degree",0.0174532925199433]],USAGE[SCOPE["Horizontal component of 3D system."],AREA["World."],BBOX[-90,-180,90,180]],ID["EPSG",4326]]'



# Built-ins
from pathlib import Path
import re
import textwrap
from typing import Literal


def extract_hdr_band_names(hdr_fp: str | Path) -> list[str]:
    ptrn = re.compile(r"band\s*names\s*=\s*{([\s\S]*)}")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        raise ValueError("Invalid .HDR format: Cannot find band names.")
    result = match.groups()[0][1:]
    return [i.strip() for i in re.split(r"\s*,\s*\n?", result)]


def extract_hdr_wavelengths(
    hdr_fp: str | Path,
) -> list[float] | Literal["Wavelengths not found."]:
    ptrn = re.compile(r"(?<=\n)wavelength\s*=\s*{([\s\S]*?)}")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        return "Wavelengths not found."
    result = match.groups()[0]
    return [float(i.strip()) for i in re.split(r"\s*,\s*\n?", result)]


def extract_hdr_bbl(hdr_fp: str | Path) -> list[int] | Literal["No BBL Found"]:
    ptrn = re.compile(r"(?<=\n)bbl\s*=\s*{([\s\S]*?)}")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        return "No BBL Found"
    result = match.groups()[0]
    return [int(i.strip()) for i in re.split(r"\s*,\s*\n?", result)]


def extract_hdr_desc(hdr_fp: str | Path) -> str:
    ptrn = re.compile(r"description\s*=\s*{([\s\S]*?)}")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        raise ValueError("Invalid .HDR format: Cannot find description.")
    result = match.groups()[0]
    return result


def replace_hdr_band_names(hdr_fp: str | Path, new_band_names: list[str]):
    ptrn = re.compile(r"band\s*names\s*=\s*{([\s\S]*)}")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        raise ValueError("Invalid .HDR format: Cannot find band names.")
    result = match.groups()[0]
    new_hdr_str = s.replace(result, ", ".join(new_band_names))
    with open(hdr_fp, "w") as dst:
        dst.write(new_hdr_str)


def replace_hdr_description(hdr_fp: str | Path, new_desc: str):
    ptrn = re.compile(r"description\s*=\s*{([\s\S]*?)}")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        raise ValueError("Invalid .HDR format: Cannot find description.")
    result = match.groups()[0]
    new_hdr_str = s.replace(result, f"\n{textwrap.fill(new_desc, width=80)}")
    with open(hdr_fp, "w") as dst:
        dst.write(new_hdr_str)


def replace_integer_field(
    hdr_fp: str | Path, field_name: str, replace_value: int
) -> None:
    ptrn = re.compile(
        rf"({field_name}\s*=\s*\d+)"
    )  # \nlines\s*=\s*\d\nbands\s*=\s*\d")
    with open(hdr_fp) as src:
        s = src.read()
    match = re.search(ptrn, s)
    if not match:
        raise ValueError(
            f"Invalid .HDR format: Cannot find integer field: {field_name}"
        )
    result = match.groups()[0]
    new_hdr_str = s.replace(result, f"{field_name} = {replace_value}")
    with open(hdr_fp, "w") as dst:
        dst.write(new_hdr_str)


def replace_shape_fields(
    hdr_fp: Path | str, *, samples: int, lines: int, bands: int
):
    replace_integer_field(hdr_fp, "samples", samples)
    replace_integer_field(hdr_fp, "lines", lines)
    replace_integer_field(hdr_fp, "bands", bands)



# Built-Ins
from pathlib import Path
from uuid import uuid4

# Dependencies
import rasterio as rio  # type: ignore

# Local Imports
from cubio.types import RasterioProfile, NumpyDType
from cubio.cube_context import CubeContext, ContextBuilder
from cubio.geotools.models import GeotransformModel
from cubio.data.crs_wkt_strings import GeographicCRS
from cubio.envi_hdr_tools import (
    extract_hdr_wavelengths,
    extract_hdr_bbl,
    extract_hdr_band_names,
    extract_hdr_desc,
)


def read_spectral_envi_file_context(fp: Path | str, name: str) -> CubeContext:
    """
    Reads the context data from an ENVI cube file representing spectral data.

    Parameters
    ----------
    fp : Path | str
        File path to ENVI cube.
    name : str
        Name of the CubeContext object.

    Returns
    -------
    CubeContext
        Relevant context data about the cube.

    Raises
    ------
    ValueError
        If the .hdr file does not contain a wavelength field.
    """

    with rio.open(fp, "r") as f:
        prf: RasterioProfile = f.profile

    hdr_fp = Path(fp).with_suffix(".hdr")
    wvls = extract_hdr_wavelengths(hdr_fp)
    desc = extract_hdr_desc(hdr_fp)
    bbl = extract_hdr_bbl(hdr_fp)
    if wvls == "Wavelengths not found.":
        raise ValueError(
            "File is not a spectral envi file. Try"
            "`read_measurement_envi_file_context()`"
        )
    if bbl == "No BBL Found":
        bbl = [1] * len(wvls)

    if prf["crs"] is None:
        crs = GeographicCRS.GCS_MOON_2000
    else:
        crs = prf["crs"]

    context_dict: ContextBuilder = {
        "name": name,
        "description": desc,
        "data_filename": Path(Path(fp).name),
        "nrows": prf["height"],
        "ncols": prf["width"],
        "nbands": prf["count"],
        "crs": crs,
        "geotransform": GeotransformModel.fromaffine(prf["transform"]),
        "hdr_off": 0,
        "data_type": NumpyDType.FLOAT32,
        "interleave": "BIL",
        "nodata": -999,
        "band_names": [f"Band{n+1}({i}nm)" for n, i in enumerate(wvls)],
        "measurement_name": "Wavelength",
        "measurement_units": "nm",
        "measurement_values": wvls,
        "bad_bands": bbl,
        "id": uuid4(),
    }

    return CubeContext.from_builder(context_dict)


def read_measurement_envi_file_context(
    fp: Path | str, name: str
) -> CubeContext:
    """
    Reads the context data from an envi file that does not represent a spectral
    dataset, but some other measurement (Location, Geometry, Parameters,
    etc...)

    Parameters
    ----------
    fp : Path | str
        File path to the ENVI cube.
    name : str
        Name of the CubeContext object.

    Returns
    -------
    CubeContext
        Resulting context data from the cube.
    """

    with rio.open(fp, "r") as f:
        prf: RasterioProfile = f.profile

    hdr_fp = Path(fp).with_suffix(".hdr")
    band_names = extract_hdr_band_names(hdr_fp)
    bbl = extract_hdr_bbl(hdr_fp)
    desc = extract_hdr_desc(hdr_fp)
    if bbl == "No BBL Found":
        bbl = [1] * len(band_names)

    if prf["crs"] is None:
        crs = GeographicCRS.GCS_MOON_2000
    else:
        crs = prf["crs"]

    context_dict: ContextBuilder = {
        "name": name,
        "description": desc,
        "data_filename": Path(Path(fp).name),
        "nrows": prf["height"],
        "ncols": prf["width"],
        "nbands": prf["count"],
        "crs": crs,
        "geotransform": GeotransformModel.fromaffine(prf["transform"]),
        "hdr_off": 0,
        "data_type": NumpyDType(prf["dtype"]),
        "interleave": "BIL",
        "nodata": -999,
        "band_names": band_names,
        "measurement_name": name,
        "measurement_units": "unitless",
        "measurement_values": [i for i in range(prf["count"])],
        "bad_bands": bbl,
        "id": uuid4(),
    }

    return CubeContext.from_builder(context_dict)



# Built-Ins
from typing import Optional
from typing_extensions import Self
from warnings import warn

# Dependencies
import xarray as xr
import numpy as np

# Local Imports
from cubio.types import LabelLike, CubeArrayFormat
from cubio.geotools.models import GeotransformModel
from cubio.cube_size_tools import get_cube_size, transpose_cube, CubeSize


class CubeMask:
    def __init__(
        self,
        *,
        shape: CubeSize,
        xy_mask: np.ndarray | xr.DataArray | None = None,
        z_mask: np.ndarray | xr.DataArray | None = None,
    ) -> None:
        self.shape = shape
        if (xy_mask is not None) and (xy_mask.dtype is not bool):
            raise ValueError("XY Mask must be of dtype: bool")
        if (z_mask is not None) and (z_mask.dtype is not bool):
            raise ValueError("Z Mask must be of dtype: bool")

        self._xymask = xy_mask
        self._zmask = z_mask

    @classmethod
    def transparent(cls, shape: CubeSize) -> Self:
        return cls(
            shape=shape,
            xy_mask=np.ones((shape.nrows, shape.ncolumns), dtype=bool),
            z_mask=np.ones(shape.nbands, dtype=bool),
        )

    @property
    def xymask(self) -> np.ndarray | xr.DataArray:
        if self._xymask is None:
            raise ValueError("XY Mask not set yet.")
        return self._xymask

    @xymask.setter
    def xymask(self, value: np.ndarray | xr.DataArray) -> None:
        if value.dtype is not bool:
            raise ValueError("XY Mask must be of type: bool")
        self._xymask = value

    @property
    def zmask(self) -> np.ndarray | xr.DataArray:
        if self._zmask is None:
            raise ValueError("XY Mask not set yet.")
        return self._zmask

    @zmask.setter
    def zmask(self, value: np.ndarray | xr.DataArray) -> None:
        if value.dtype is not bool:
            raise ValueError("Z Mask must be of type: bool")
        self._zmask = value


class CubeData:
    def __init__(
        self,
        name: str,
        format: CubeArrayFormat,
        *,
        x_labels: Optional[LabelLike] = None,
        y_labels: Optional[LabelLike] = None,
        z_labels: Optional[LabelLike] = None,
        x_name: str = "XAxis",
        y_name: str = "YAxis",
        z_name: str = "ZAxis",
        geotransform: Optional[GeotransformModel] = None,
    ) -> None:
        self.name: str = name
        self.fmt: CubeArrayFormat = format
        self._gtrans = geotransform

        self._array: xr.DataArray | None = None

        self._xlbl = x_labels
        self._ylbl = y_labels
        self._zlbl = z_labels

        self.xname = x_name
        self.yname = y_name
        self.zname = z_name

        if self._gtrans is not None:
            if (self._xlbl is not None) | (self._ylbl is not None):
                warn(
                    "A geotransform was provided, so any provided x and y"
                    f" labels will be overwritten for {self.name}"
                )
            self.xname = "Longitude"
            self.yname = "Latitude"

            self._mask: CubeMask | None = None

    @property
    def mask(self) -> CubeMask:
        if self._mask is None:
            return CubeMask.transparent(self.shape)
        return self._mask

    @mask.setter
    def mask(self, value: CubeMask) -> None:
        self._mask = value

    @property
    def shape(self) -> CubeSize:
        if (
            (self._ylbl is None)
            or (self._xlbl is None)
            or (self._zlbl is None)
        ):
            raise ValueError("Cube Data is not set yet.")
        return CubeSize(
            nrows=len(self._ylbl),
            ncolumns=len(self._xlbl),
            nbands=len(self._zlbl),
        )

    @property
    def array(self) -> xr.DataArray:
        if self._array is None:
            raise NotImplementedError(
                f"A data array has not been set for {self.name}."
            )
        return self._array

    @array.setter
    def array(self, value: xr.DataArray) -> None:
        s = get_cube_size(value, self.fmt)

        # Setting label values as indices if none are provided.
        if self._xlbl is None:
            self._xlbl = np.arange(0, s.ncolumns)
        if self._ylbl is None:
            self._ylbl = np.arange(0, s.nrows)
        if self._zlbl is None:
            self._zlbl = np.arange(0, s.nbands)

        # X and Y labels should be replaced with coordinates, if possible.
        if self.geotransform is not None:
            self._xlbl, self._ylbl = self.geotransform.generate_coords(
                width=s.ncolumns, height=s.nrows
            )

        dims: tuple[str, str, str]
        if self.fmt == "BIL":
            dims = (self.yname, self.zname, self.xname)
        elif self.fmt == "BIP":
            dims = (self.yname, self.xname, self.zname)
        elif self.fmt == "BSQ":
            dims = (self.zname, self.xname, self.yname)
        else:
            raise ValueError("Invalid axis names.")

        self._array = xr.DataArray(
            value.data,
            coords={
                self.xname: self._xlbl,
                self.yname: self._ylbl,
                self.zname: self._zlbl,
            },
            dims=dims,
        )

    @property
    def geotransform(self) -> GeotransformModel:
        if self._gtrans is None:
            raise ValueError("Geotransform has not been set.")
        return self._gtrans

    @geotransform.setter
    def geotransform(self, value: GeotransformModel) -> None:
        self._gtrans = value
        self.array = self.array

    def transpose_to(self, format: CubeArrayFormat) -> None:
        old_format = self.fmt
        self.fmt = format
        new_arr = transpose_cube(old_format, format, self.array)
        self.array = new_arr

    def transpose_to_rasterio(self) -> xr.DataArray:
        return transpose_cube(self.fmt, "RASTERIO", self.array)



# Built-ins
from typing import TypedDict, Literal
from typing_extensions import Self
from tempfile import NamedTemporaryFile
from pathlib import Path
import shutil
from uuid import UUID, uuid4

# Dependencies
import numpy as np
from pydantic import (
    BaseModel,
    field_serializer,
    model_validator,
    Field,
    field_validator,
    PrivateAttr,
)
import rasterio as rio  # type: ignore
import xarray as xr

# Local Imports
from cubio.types import (
    NumpyDType,
    CubeArrayFormat,
    is_valid_cubearrayformat,
    suffix_to_format_map,
    RasterioProfile,
    ImageSuffix,
    is_valid_image_suffix,
    image_suffix_priority,
)
from cubio.geotools.models import GeotransformModel
from cubio.envi_hdr_tools import (
    replace_hdr_band_names,
    replace_hdr_description,
    replace_shape_fields,
)
from cubio.cube_size_tools import CubeSize
from cubio.cube_data import CubeData


class ContextBuilder(TypedDict):
    name: str
    description: str
    data_filename: Path
    ncols: int
    nrows: int
    nbands: int
    hdr_off: int
    data_type: NumpyDType
    interleave: CubeArrayFormat
    crs: str
    geotransform: GeotransformModel
    band_names: list[str]
    nodata: float | int
    measurement_name: str
    measurement_units: str
    measurement_values: list[float]
    bad_bands: list[int]
    id: UUID


class CubeContext(BaseModel):
    name: str = Field(..., description="A short name for the cube.")
    description: str = Field(
        ..., description="A description of the data cube."
    )
    data_filename: Path = Field(
        ...,
        description="Filename of the data that is described by this file. It"
        " should be in the same directory as this file.",
    )
    ncols: int = Field(..., description="Number of columns in the data cube.")
    nrows: int = Field(..., description="Number of rows in the data cube.")
    nbands: int = Field(..., description="Number of bands in the data cube.")
    hdr_off: int = Field(default=0, description="Header offset in bytes.")
    data_type: NumpyDType = Field(
        ..., description="The data type of the cube."
    )
    interleave: CubeArrayFormat = Field(
        ...,
        description=(
            "The interleave format of the cube. Either BIL, BIP, or BSQ."
        ),
    )
    crs: str = Field(
        ..., description="Coordinate reference system of the cube."
    )
    geotransform: GeotransformModel = Field(
        ..., description="Geotransform model for the cube."
    )
    band_names: list[str] = Field(default_factory=list)
    nodata: float | int = Field(default=-999, description="The nodata value.")
    measurement_name: str = Field(
        default="Measurement",
        description="Name that describes the nature of the measurement along "
        "the cube z-axis.",
    )
    measurement_units: str = Field(
        default="unitless", description="The measurement units."
    )
    measurement_values: list[float] = Field(
        default_factory=list, description="The measurement values."
    )
    bad_bands: list[int] = Field(
        default_factory=list, description="List of bad band flags."
    )
    id: UUID = Field(
        default_factory=uuid4, description="Unique ID of the cube object."
    )
    _retrieval_path: Path | Literal["NoRetrieval"] = PrivateAttr(
        default="NoRetrieval"
    )

    @property
    def shape(self) -> CubeSize:
        return CubeSize(
            nrows=self.nrows, ncolumns=self.ncols, nbands=self.nbands
        )

    @property
    def shape_tuple(self) -> tuple[int, int, int]:
        return self.shape.as_tuple(self.interleave)

    @property
    def builder(self) -> ContextBuilder:
        _builder: ContextBuilder = {
            "name": self.name,
            "description": self.description,
            "data_filename": self.data_filename,
            "ncols": self.ncols,
            "nrows": self.nrows,
            "nbands": self.nbands,
            "hdr_off": self.hdr_off,
            "data_type": self.data_type,
            "interleave": self.interleave,
            "crs": self.crs,
            "geotransform": self.geotransform,
            "band_names": self.band_names,
            "nodata": self.nodata,
            "measurement_name": self.measurement_name,
            "measurement_units": self.measurement_units,
            "measurement_values": self.measurement_values,
            "bad_bands": self.bad_bands,
            "id": self.id,
        }
        return _builder

    @classmethod
    def from_builder(cls, builder_dict: ContextBuilder) -> Self:
        return cls(**builder_dict)

    @classmethod
    def from_json(cls, savefp: str | Path) -> Self:
        """Convenience method for reading in the model from json file."""
        with open(savefp, "r") as f:
            json = f.read()
        valid_model = cls.model_validate_json(json)
        valid_model._retrieval_path = Path(savefp)
        return valid_model

    @field_serializer("interleave", mode="plain")
    def lowercase(self, value: CubeArrayFormat) -> str:
        return value.lower()

    @field_validator("interleave", mode="before")
    @classmethod
    def uppercase(cls, value: str) -> CubeArrayFormat:
        ustr = value.upper()
        if is_valid_cubearrayformat(ustr):
            return ustr
        else:
            raise ValueError(f"Invalid interleave: {ustr}")

    @model_validator(mode="after")
    def set_default_bbl(self) -> Self:
        current_bbl = self.bad_bands
        if len(current_bbl) == 0:
            self.bad_bands = [1] * len(self.measurement_values)
        else:
            self.bad_bands = current_bbl

        if len(self.bad_bands) != self.nbands:
            print(self.bad_bands)
            raise ValueError(
                "Length of bad band list must match the number of bands."
            )

        return self

    @model_validator(mode="after")
    def check_measurement_values(self) -> Self:
        if len(self.measurement_values) != self.nbands:
            raise ValueError(
                "Length of measurement values must match the number of bands."
            )
        return self

    def write_envi_hdr(
        self, dst: Path | str | None = None, use_image_name: bool = True
    ) -> None:
        """
        Writes the model to a properly formatted ENVI header file.

        Parameters
        ----------
        dst : Path | str | None, optional
            Destination file path, by default None.

        Notes
        -----
        The file suffix is set internally, so the dst file does not need to
        include one.
        """

        # A temporary file should be used if no path is provided.
        if dst is None:
            tf = NamedTemporaryFile(suffix=".hdr", delete=False)
            savefp = Path(tf.name)
        else:
            savefp = Path(dst)

        # Adding suffix based on the use_image_name option.
        if use_image_name:
            if savefp.is_dir():
                savefp = Path(savefp, self.data_filename).with_suffix(".hdr")
            else:
                savefp = savefp.with_name(str(self.data_filename)).with_suffix(
                    ".hdr"
                )
        else:
            savefp = savefp.with_suffix(".hdr")

        # Need a temporary image and hdr file to write to. These will be empty.
        temp_img = savefp.with_name(f"_{savefp.name}").with_suffix(".img")
        temp_hdr = temp_img.with_suffix(".hdr")

        # Need a rasterio profile to do a proper write.
        envi_profile: RasterioProfile = {
            "driver": "ENVI",
            "width": 1,
            "height": 1,
            "count": 1,
            "dtype": self.data_type,
            "crs": self.crs,
            "transform": self.geotransform.toaffine(),
            "interleave": self.interleave,
            "nodata": -999,
        }

        # A blank image write is needed. We just use the .hdr file.
        with rio.open(temp_img, "w", **envi_profile) as f:
            f.write(np.empty((1, 1, 1)))

        temp_img.unlink()  # Discarding temporary image.

        # Modifying rasterio-written .hdr file.
        # These are existing fields.
        replace_shape_fields(
            temp_hdr,
            samples=self.ncols,
            lines=self.nrows,
            bands=self.nbands,
        )
        replace_hdr_band_names(temp_hdr, self.band_names)
        replace_hdr_description(temp_hdr, self.description)

        with open(temp_hdr, "a") as f:
            # These are all new fields.
            f.write(f"data ignore value = {self.nodata}\n")
            f.write(f"wavelenth units = {self.measurement_units}\n")
            f.write(
                "wavelength = "
                f"{{{",".join([str(i) for i in self.measurement_values])}}}\n"
            )
            f.write(
                "bbl = " f"{{{",".join([str(i) for i in self.bad_bands])}}}\n"
            )
        # Renaming the temporary hdr that was created with the temp image.
        shutil.move(temp_hdr, temp_hdr.with_name(temp_hdr.name[1:]))

        # Writing out the pydantic model in json form for easy reading.
        self.write_json(savefp)

    def write_json(self, savefp: str | Path) -> None:
        """Convenience function for dumping the model to a json file."""
        json_str = self.model_dump_json(indent=2)
        with open(Path(savefp).with_suffix(".json"), "w") as f:
            f.write(json_str)

    def lazy_load_data(self, search_dir: str | Path | None = None) -> CubeData:
        load_from: Path
        if (search_dir is not None) and (
            self._retrieval_path == "NoRetrieval"
        ):
            load_from = Path(search_dir)
        elif (self._retrieval_path != "NoRetrieval") and (search_dir is None):
            load_from = self._retrieval_path.parent
        else:
            raise ValueError(
                "If `savefp` is specified, the object must not have been "
                "validated from disk."
            )

        if not load_from.is_dir():
            raise ValueError(f"Search directory does not exist: {load_from}")

        dat = CubeData(
            self.name,
            self.interleave,
            z_labels=self.measurement_values,
            z_name=self.measurement_name,
            geotransform=self.geotransform,
        )
        candidate_image_data_files: list[tuple[Path, ImageSuffix]] = []
        for i in load_from.iterdir():
            if i.stem == str(self.data_filename.stem):
                suff = i.suffix.lower()
                if is_valid_image_suffix(suff):
                    candidate_image_data_files.append((i, suff))

        candidate_image_data_files.sort(
            key=lambda item: image_suffix_priority[item[1]]
        )

        image_data_file = candidate_image_data_files[0][0]

        interleave_test = suffix_to_format_map.get(image_data_file.suffix)

        if (interleave_test is not None) and (
            interleave_test != self.interleave
        ):
            raise ValueError(
                f"Loaded data ({image_data_file}) does not match the"
                f"registered interleave ({self.interleave})"
            )

        if image_data_file.suffix.lower() in [".img", ".bsq", ".bil", ".bip"]:
            mmap = np.memmap(
                image_data_file,
                dtype=np.dtype(self.data_type),
                shape=self.shape_tuple,
            )
            dat.array = xr.DataArray(mmap)
        elif image_data_file.suffix.lower() == ".hdf5":
            raise NotImplementedError("HDF5 file not implemented yet.")
        elif image_data_file.suffix.lower() == ".zarr":
            arr = xr.open_dataarray(image_data_file, engine="zarr")
            dat.yname = str(arr.dims[0])
            dat.xname = str(arr.dims[1])
            dat.zname = str(arr.dims[2])
            dat.array = arr

        return dat



# Built-Ins
from dataclasses import dataclass
from typing_extensions import Self

# Dependencies
import numpy as np
from pyresample.geometry import AreaDefinition, SwathDefinition
from pyresample import kd_tree
import xarray as xr
from tqdm import tqdm

# Local
from .models import BoundingBoxModel
from cubio.geotools.models import GeotransformModel, PointModel

MOON_RADIUS = 1737400
MOON_M_PER_DEG = np.pi * MOON_RADIUS / 180


@dataclass
class ProjectionDefinition:
    area_name: str
    proj_name: str
    description: str
    proj4_str: str
    crs_wkt_str: str


@dataclass
class PixelResolution:
    pix_per_deg: float
    deg_per_pix: float
    m_per_pix: float  # Estimation

    @classmethod
    def from_array(cls, max_lat: float, min_lat: float, height: int) -> Self:
        lat_height = max_lat - min_lat
        dpp: float = lat_height / height  # Degrees per pixel
        return cls(
            pix_per_deg=1 / dpp,
            deg_per_pix=dpp,
            m_per_pix=MOON_M_PER_DEG * dpp,
        )


def georeference_satellite_swath(
    satellite_data: np.ndarray | xr.DataArray,
    longitude_backplane: np.ndarray | xr.DataArray,
    latitude_backplane: np.ndarray | xr.DataArray,
    proj: ProjectionDefinition,
    extent: BoundingBoxModel,
) -> tuple[np.ndarray, GeotransformModel]:
    # ---- Automatically detecting resolution ----
    max_lat = float(latitude_backplane[0, :].max())
    min_lat = float(latitude_backplane[-1, :].min())
    res = PixelResolution.from_array(
        max_lat=max_lat, min_lat=min_lat, height=satellite_data.shape[0]
    )

    # ---- Defining Swath ----
    swath = SwathDefinition(lons=longitude_backplane, lats=latitude_backplane)

    # ---- Defining Projection Area ---
    area_height = (extent.top - extent.bottom) * res.pix_per_deg
    area_width = (extent.right - extent.left) * res.pix_per_deg

    area = AreaDefinition(
        proj.area_name,
        proj.description,
        proj.proj_name,
        proj.proj4_str,
        area_width,
        area_height,
        area_extent=extent.as_extent(mode="BottomLeft"),
    )

    # ---- Resampling ----
    resampled_data: np.ndarray
    if satellite_data.ndim == 2:
        resampled_data = kd_tree.resample_nearest(
            swath,
            satellite_data,
            area,
            res.m_per_pix * 4,
            epsilon=0.5,  # type: ignore
            fill_value=np.nan,  # type: ignore
        )
    elif satellite_data.ndim == 3:
        first_band_resamp = kd_tree.resample_nearest(
            swath,
            satellite_data[:, :, 0],
            area,
            res.m_per_pix * 3,
            epsilon=0.5,  # type: ignore
            fill_value=np.nan,  # type: ignore
        )
        if not isinstance(first_band_resamp, np.ndarray):
            raise ValueError("Invalid first band resampling.")
        resampled_data = np.empty(
            (*first_band_resamp.shape, satellite_data.shape[2])
        )
        resampled_data[:, :, 0] = first_band_resamp
        for band in tqdm(
            range(1, satellite_data.shape[2]),
            desc="Resampling bands...",
            total=satellite_data.shape[2] - 1,
        ):
            resampled_data[:, :, band] = kd_tree.resample_nearest(
                swath,
                satellite_data[:, :, band],
                area,
                res.m_per_pix * 3,
                epsilon=0.5,  # type: ignore
                fill_value=np.nan,  # type: ignore
            )
    else:
        raise ValueError(
            "Satellite data has an invalid number of dimensions:"
            f" {satellite_data.ndim}."
        )

    gtrans = GeotransformModel(
        upperleft=PointModel(x=extent.top_left.x, y=extent.top_left.y),
        xres=res.deg_per_pix,
        yres=-res.deg_per_pix,
        col_rotation=0,
        row_rotation=0,
    )

    return resampled_data, gtrans



# Dependencies
import numpy as np
from scipy.spatial import Delaunay
import xarray as xr

# Local
from .models.gcp_model import GCPGroup


def generate_latlong(
    col_gcps: np.ndarray,
    row_gcps: np.ndarray,
    lon_gcps: np.ndarray,
    lat_gcps: np.ndarray,
    height: int,
    width: int,
) -> np.ndarray:
    # GCP arrays
    pix = np.column_stack([col_gcps, row_gcps])
    latlon = np.column_stack([lat_gcps, lon_gcps])

    tri = Delaunay(pix)

    # Full pixel grid
    jj, ii = np.meshgrid(np.arange(height), np.arange(width), indexing="ij")
    pts = np.column_stack([ii.ravel(), jj.ravel()])

    simplices = tri.find_simplex(pts)
    T = tri.transform[simplices]

    bary = np.einsum("ijk,ik->ij", T[:, :2, :], pts - T[:, 2])
    bary = np.c_[bary, 1 - bary.sum(axis=1)]

    latlon_dense: np.ndarray = np.sum(
        latlon[tri.simplices[simplices]] * bary[..., None], axis=1
    )

    return latlon_dense.reshape((height, width, 2))


def latlong_from_gcp_group(
    gcp_group: GCPGroup, base_image: np.ndarray | xr.DataArray
) -> np.ndarray:
    return generate_latlong(
        gcp_group.col_pixels,
        gcp_group.row_pixels,
        gcp_group.map_x,
        gcp_group.map_y,
        base_image.shape[0],
        base_image.shape[1],
    )



# Built-Ins
from pathlib import Path
from copy import copy

# Dependencies
import numpy as np
import xarray as xr

# Local
from .models import BoundingBoxModel, GCPGroup
from .georeference_satellite_swath import (
    georeference_satellite_swath,
    ProjectionDefinition,
)
from .generate_geoloc_backplane import latlong_from_gcp_group
from cubio.cube_context import CubeContext
from cubio.cube_data import CubeData
from cubio.geotools.models import ImageOffset, GeotransformModel


def georeference_image(
    cubio_json_file: Path | str,
    gcps_file: Path | str,
    prj_definition: ProjectionDefinition,
    unref_cube_array: np.ndarray | None = None,
    georef_extent: BoundingBoxModel | None = None,
    new_gcps_offset: ImageOffset | None = None,
) -> tuple[np.ndarray, GeotransformModel]:
    # ---- Reading Image Context and Lazy Loading ----
    unreferenced_context = CubeContext.from_json(cubio_json_file)
    if unref_cube_array is None:
        unref_cube = unreferenced_context.lazy_load_data()
        unref_cube.transpose_to("BIP")
    elif unref_cube_array is not None:
        unref_cube = CubeData("unref_cube", format="BIP")
        unref_cube.array = xr.DataArray(unref_cube_array)

    # ---- Cropping to extent of GCP group ----
    gcp_group = GCPGroup.from_gcps_file(gcps_file)
    if new_gcps_offset:
        gcp_group.adjust_offset(new_gcps_offset)
    offset_cube: np.ndarray = np.array(
        gcp_group.offset.crop_image(unref_cube.array)
    )

    # ---- Creating Lat/Long Backplane ----
    latlongarr = latlong_from_gcp_group(gcp_group, offset_cube)
    long = latlongarr[:, :, 1]
    lat = latlongarr[:, :, 0]

    # ---- Automatically obtaining extent, if one is not provided ----
    if georef_extent is None:
        ext = BoundingBoxModel(
            name="auto",
            top=lat.max() - 0.2,
            bottom=lat.min() + 0.2,
            left=long.min() + 0.2,
            right=long.max() - 0.2,
        )
    else:
        ext = georef_extent

    # ---- Resampling Image ----
    resamp_img, gtrans = georeference_satellite_swath(
        satellite_data=offset_cube,
        longitude_backplane=long,
        latitude_backplane=lat,
        proj=prj_definition,
        extent=ext,
    )

    return resamp_img, gtrans


def save_georeference(
    cubio_json_file: Path | str,
    resamp_img: np.ndarray,
    gtrans: GeotransformModel,
    prj_definition: ProjectionDefinition,
    unreferenced_context: CubeContext,
    save_directory: Path | str | None = None,
    new_name: str | None = None,
    new_description: str | None = None,
    save_new_context: bool = True,
):
    # ---- Export Main Array to Disk ----
    xcoord, ycoord = gtrans.generate_coords(
        height=resamp_img.shape[0], width=resamp_img.shape[1]
    )
    meas_name = unreferenced_context.measurement_name
    meas_vals = unreferenced_context.measurement_values
    darray = xr.DataArray(
        resamp_img,
        coords={
            "Latitude": ycoord,
            "Longitude": xcoord,
            meas_name: meas_vals,
        },
        dims=("Latitude", "Longitude", meas_name),
    )
    if save_directory is None:
        src_fp = Path(cubio_json_file)
        dst_fp = src_fp.with_name(f"{src_fp.stem}_georef")
        save_fp = dst_fp.with_suffix(".zarr")
    else:
        dst_fp = Path(save_directory, unreferenced_context.name)
        save_fp = dst_fp.with_suffix(".zarr")
    darray.to_zarr(save_fp, mode="w", consolidated=True, zarr_format=2)

    referenced_context = copy(unreferenced_context.builder)
    referenced_context.update(
        {
            "crs": prj_definition.crs_wkt_str,
            "geotransform": gtrans,
            "nrows": resamp_img.shape[0],
            "ncols": resamp_img.shape[1],
            "data_filename": Path(save_fp.stem),
        }
    )

    if new_name is not None:
        referenced_context["name"] = new_name
    if new_description is not None:
        referenced_context["description"] = new_description

    georef_context = CubeContext.from_builder(referenced_context)

    if save_new_context:
        georef_context.write_json(save_fp.with_suffix(".json"))

    return georef_context



from pydantic import BaseModel, Field
import fiona  # type: ignore
from shapely.geometry import Polygon, mapping  # type: ignore
from pathlib import Path
from uuid import UUID, uuid4
from typing import Literal


def todms(
    dd: float, coord_type: Literal["lon", "lat"], precision: int = 0
) -> str:
    """Convert decimal degrees (-180..180) to DMS with N/S/E/W.

    Args:
        dd: Decimal degrees value.
        coord_type: 'lat' for latitude (N/S) or 'lon' for longitude (E/W').
        precision: Number of decimal places for seconds.

    Returns:
        A string like "123°45'56.78"W" or "12°34'56.78"N".
    """
    if coord_type not in ("lat", "lon"):
        raise ValueError("coord_type must be 'lat' or 'lon'")

    hemi = None
    if coord_type == "lat":
        hemi = "N" if dd >= 0 else "S"
    else:
        hemi = "E" if dd >= 0 else "W"

    total_seconds = round(abs(dd) * 3600, precision)
    degrees = int(total_seconds // 3600)
    remainder = total_seconds - degrees * 3600
    minutes = int(remainder // 60)
    seconds = remainder - minutes * 60

    return f"{degrees}°{minutes}'{seconds:.{precision}f}\"{hemi}"


class Point(BaseModel):
    x: float
    y: float

    def as_tuple(self) -> tuple[float, float]:
        return (self.x, self.y)


class BoundingBoxModel(BaseModel):
    left: float = Field(
        ..., description="The left (minimum x) coordinate of the bounding box."
    )
    bottom: float = Field(
        ...,
        description="The bottom (minimum y) coordinate of the bounding box.",
    )
    right: float = Field(
        ...,
        description="The right (maximum x) coordinate of the bounding box.",
    )
    top: float = Field(
        ..., description="The top (maximum y) coordinate of the bounding box."
    )
    name: str = Field(
        ..., description="A name or identifier for the bounding box."
    )
    id: UUID = Field(
        default_factory=uuid4,
        description="A unique identifier for the bounding box.",
    )

    @property
    def top_left(self) -> Point:
        return Point(x=self.left, y=self.top)

    @property
    def top_right(self) -> Point:
        return Point(x=self.right, y=self.top)

    @property
    def bottom_left(self) -> Point:
        return Point(x=self.left, y=self.bottom)

    @property
    def bottom_right(self) -> Point:
        return Point(x=self.right, y=self.bottom)

    @property
    def shapely_polygon(self) -> Polygon:
        return Polygon(
            [
                self.top_left.as_tuple(),
                self.top_right.as_tuple(),
                self.bottom_right.as_tuple(),
                self.bottom_left.as_tuple(),
            ]
        )

    def __str__(self) -> str:
        space = "    "
        return (
            f"{self.name}:\n"
            f"Decimal Degrees:\n"
            f"{space}Left Boundary: {self.left}\n"
            f"{space}Bottom Boundary: {self.bottom}\n"
            f"{space}Right Boundary: {self.right}\n"
            f"{space}Top Boundary: {self.top}\n"
            f"\nDegrees Minutes Seconds: \n"
            f"{space}Left Boundary: {todms(self.left, coord_type='lon')}\n"
            f"{space}Bottom Boundary: {todms(self.bottom, coord_type='lat')}\n"
            f"{space}Right Boundary: {todms(self.right, coord_type='lon')}\n"
            f"{space}Top Boundary: {todms(self.top, coord_type='lat')}\n"
        )

    def as_csv_row(self) -> str:
        return (
            f"{self.name}, {self.left}, {self.bottom}, {self.right}, "
            f"{self.top}\n"
        )

    def as_extent(
        self, mode: Literal["TopLeft", "BottomLeft"] = "BottomLeft"
    ) -> list[float]:
        if mode == "BottomLeft":
            return [
                self.bottom_left.x,
                self.bottom_left.y,
                self.top_right.x,
                self.top_right.y,
            ]
        elif mode == "TopLeft":
            return [
                self.top_left.x,
                self.top_left.y,
                self.bottom_right.x,
                self.bottom_right.y,
            ]


def to_shapefile(
    bbox: BoundingBoxModel | list[BoundingBoxModel],
    crs: str,
    dst_fp: Path | str,
) -> None:
    schema = {
        "geometry": "Polygon",
        "properties": {"name": "str"},
    }
    fiona_config = {
        "crs": crs,
        "driver": "ESRI Shapefile",
        "schema": schema,
    }
    bbox_list: list[BoundingBoxModel]
    if isinstance(bbox, list):
        bbox_list = bbox
    else:
        bbox_list = [bbox]

    if Path(dst_fp).is_dir():
        print(f"Writing shapefiles to directory: {dst_fp}")
        for i in bbox_list:
            save = Path(dst_fp, i.name).with_suffix(".shp")
            with fiona.open(save, "w", **fiona_config) as c:
                c.write(
                    {
                        "geometry": mapping(i.shapely_polygon),
                        "properties": {"name": i.name},
                    }
                )
    else:
        save = Path(dst_fp).with_suffix(".shp")
        print(f"Writing to single shapefile: {save}")
        with fiona.open(save, "w", **fiona_config) as c:
            for i in bbox_list:
                c.write(
                    {
                        "geometry": mapping(i.shapely_polygon),
                        "properties": {"name": i.name},
                    }
                )


def to_csv(bbox: list[BoundingBoxModel], dst_fp: Path | str) -> None:
    with open(Path(dst_fp).with_suffix(".csv"), "w") as f:
        f.write(
            "Region, Left_Longitude, Bottom_Latitude, Right_Longitude, "
            "Top_Latitude\n"
        )
        for i in bbox:
            f.write(i.as_csv_row())
